@@ -1,6 +1,6 @@
 /* enchant
  * Copyright (C) 2003, 2004 Dom Lachowicz
- * Copyright (C) 2017-2021 Reuben Thomas <rrt@sc3d.org>
+ * Copyright (C) 2016-2023 Reuben Thomas <rrt@sc3d.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -112,7 +112,7 @@ enchant_ensure_dir_exists (const char* dir)
 		}
 }
 
-char *
+_GL_ATTRIBUTE_MALLOC char *
 enchant_get_user_config_dir (void)
 {
 	const gchar * env = g_getenv("ENCHANT_CONFIG_DIR");
@@ -266,42 +266,31 @@ enchant_session_new_with_pwl (EnchantProvider * provider,
 }
 
 static EnchantSession *
-_enchant_session_new (EnchantProvider *provider, const char * const user_config_dir,
-		      const char * const lang, gboolean fail_if_no_pwl)
-{
-	if (!user_config_dir || !lang)
-		return NULL;
-
-	char *filename = g_strdup_printf ("%s.dic", lang);
-	char *dic = g_build_filename (user_config_dir, filename, NULL);
-	g_free (filename);
-
-	filename = g_strdup_printf ("%s.exc", lang);
-	char *excl = g_build_filename (user_config_dir, filename, NULL);
-	g_free (filename);
-
-	EnchantSession * session = enchant_session_new_with_pwl (provider, dic, excl, lang, fail_if_no_pwl);
-
-	g_free (dic);
-	g_free (excl);
-
-	return session;
-}
-
-static EnchantSession *
-enchant_session_new (EnchantProvider *provider, const char * const lang)
+enchant_session_new (EnchantProvider *provider, const char * const lang, const char * pwl)
 {
 	char *user_config_dir = enchant_get_user_config_dir ();
 
 	EnchantSession * session = NULL;
-	session = _enchant_session_new (provider, user_config_dir, lang, TRUE);
+	if (!user_config_dir || !lang)
+		return NULL;
+	enchant_ensure_dir_exists (user_config_dir);
 
-	if (session == NULL && user_config_dir != NULL)
-		{
-			enchant_ensure_dir_exists (user_config_dir);
-			session = _enchant_session_new (provider, user_config_dir, lang, FALSE);
-		}
+	char *filename, *excl = NULL, *alloc_pwl = NULL;
+	if (pwl == NULL) {
+		filename = g_strdup_printf ("%s.dic", lang);
+		pwl = alloc_pwl = g_build_filename (user_config_dir, filename, NULL);
+		g_free (filename);
 
+		filename = g_strdup_printf ("%s.exc", lang);
+		excl = g_build_filename (user_config_dir, filename, NULL);
+		g_free (filename);
+	}
+
+	// Insist that PWL be opened only if we supplied it explicitly.
+	session = enchant_session_new_with_pwl (provider, pwl, excl, lang, alloc_pwl == NULL);
+
+	g_free (alloc_pwl);
+	g_free (excl);
 	g_free (user_config_dir);
 
 	return session;
@@ -323,30 +312,6 @@ enchant_session_remove (EnchantSession * session, const char * const word, size_
 	g_hash_table_insert (session->session_exclude, key, GINT_TO_POINTER(TRUE));
 }
 
-static void
-enchant_session_add_personal (EnchantSession * session, const char * const word, size_t len)
-{
-	enchant_pwl_add(session->personal, word, len);
-}
-
-static void
-enchant_session_remove_personal (EnchantSession * session, const char * const word, size_t len)
-{
-	enchant_pwl_remove(session->personal, word, len);
-}
-
-static void
-enchant_session_add_exclude (EnchantSession * session, const char * const word, size_t len)
-{
-	enchant_pwl_add(session->exclude, word, len);
-}
-
-static void
-enchant_session_remove_exclude (EnchantSession * session, const char * const word, size_t len)
-{
-	enchant_pwl_remove(session->exclude, word, len);
-}
-
 /* a word is excluded if it is in the exclude dictionary or in the session exclude list
  *  AND the word has not been added to the session include list
  */
@@ -510,8 +475,8 @@ enchant_dict_suggest (EnchantDict * dict, const char *const word, ssize_t len, s
 	g_return_val_if_fail (len, NULL);
 	g_return_val_if_fail (g_utf8_validate(word, len, NULL), NULL);
 
-	size_t n_dict_suggs = 0, n_pwl_suggs = 0, n_suggsT = 0;
-	char **dict_suggs = NULL, **pwl_suggs = NULL, **suggsT;
+	size_t n_dict_suggs = 0, n_suggsT = 0;
+	char **dict_suggs = NULL, **suggsT;
 
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 	enchant_session_clear_error (session);
@@ -529,34 +494,18 @@ enchant_dict_suggest (EnchantDict * dict, const char *const word, ssize_t len, s
 				}
 		}
 
-	/* Check for suggestions from personal dictionary */
-	if (session->personal)
-		{
-			pwl_suggs = enchant_pwl_suggest(session->personal, word, len, dict_suggs, &n_pwl_suggs);
-			if (pwl_suggs)
-				{
-					suggsT = enchant_dict_get_good_suggestions(dict, pwl_suggs, n_pwl_suggs, &n_suggsT);
-					enchant_free_string_list (pwl_suggs);
-					pwl_suggs = suggsT;
-					n_pwl_suggs = n_suggsT;
-				}
-		}
-
 	/* Clone suggestions, if any */
 	char **suggs = NULL;
-	size_t n_suggs = n_pwl_suggs + n_dict_suggs;
+	size_t n_suggs = n_dict_suggs;
 	if (n_suggs > 0)
 		{
 			suggs = g_new0 (char *, n_suggs + 1);
 			n_suggs = 0;
 			if (dict_suggs != NULL)
 				n_suggs = enchant_dict_merge_suggestions(suggs, n_suggs, dict_suggs, n_dict_suggs);
-			if (pwl_suggs != NULL)
-				n_suggs = enchant_dict_merge_suggestions(suggs, n_suggs, pwl_suggs, n_pwl_suggs);
 		}
 
 	g_strfreev(dict_suggs);
-	g_strfreev(pwl_suggs);
 
 	if (out_n_suggs)
 		*out_n_suggs = n_suggs;
@@ -564,6 +513,24 @@ enchant_dict_suggest (EnchantDict * dict, const char *const word, ssize_t len, s
 	return suggs;
 }
 
+static void
+enchant_dict_session_add (EnchantDict * dict, const char *const word, ssize_t len)
+{
+	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
+	enchant_session_add (session, word, len);
+	if (dict->add_to_session)
+		(*dict->add_to_session) (dict, word, len);
+}
+
+static void
+enchant_dict_session_remove (EnchantDict * dict, const char *const word, ssize_t len)
+{
+	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
+	enchant_session_remove (session, word, len);
+	if (dict->remove_from_session)
+		(*dict->remove_from_session) (dict, word, len);
+}
+
 void
 enchant_dict_add (EnchantDict * dict, const char *const word, ssize_t len)
 {
@@ -578,11 +545,10 @@ enchant_dict_add (EnchantDict * dict, const char *const word, ssize_t len)
 
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 	enchant_session_clear_error (session);
-	enchant_session_add_personal (session, word, len);
-	enchant_session_remove_exclude (session, word, len);
 
-	if (dict->add_to_personal)
-		(*dict->add_to_personal) (dict, word, len);
+	enchant_dict_session_add (dict, word, len);
+	enchant_pwl_add(session->personal, word, len);
+	enchant_pwl_remove(session->exclude, word, len);
 }
 
 void
@@ -600,9 +566,7 @@ enchant_dict_add_to_session (EnchantDict * dict, const char *const word, ssize_t
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 	enchant_session_clear_error (session);
 
-	enchant_session_add (session, word, len);
-	if (dict->add_to_session)
-		(*dict->add_to_session) (dict, word, len);
+	enchant_dict_session_add (dict, word, len);
 }
 
 int
@@ -638,11 +602,9 @@ enchant_dict_remove (EnchantDict * dict, const char *const word, ssize_t len)
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 	enchant_session_clear_error (session);
 
-	enchant_session_remove_personal (session, word, len);
-	enchant_session_add_exclude(session, word, len);
-
-	if (dict->add_to_exclude)
-		(*dict->add_to_exclude) (dict, word, len);
+	enchant_dict_session_remove (dict, word, len);
+	enchant_pwl_remove(session->personal, word, len);
+	enchant_pwl_add(session->exclude, word, len);
 }
 
 void
@@ -660,7 +622,7 @@ enchant_dict_remove_from_session (EnchantDict * dict, const char *const word, ss
 	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
 	enchant_session_clear_error (session);
 
-	enchant_session_remove (session, word, len);
+	enchant_dict_session_remove (dict, word, len);
 }
 
 int
@@ -681,33 +643,12 @@ enchant_dict_is_removed (EnchantDict * dict, const char *const word, ssize_t len
 	return enchant_session_exclude (session, word, len);
 }
 
+/* Stub for obsolete API. */
 void
-enchant_dict_store_replacement (EnchantDict * dict,
-				const char *const mis, ssize_t mis_len,
-				const char *const cor, ssize_t cor_len)
+enchant_dict_store_replacement (EnchantDict * dict _GL_UNUSED,
+				const char *const mis _GL_UNUSED, ssize_t mis_len _GL_UNUSED,
+				const char *const cor _GL_UNUSED, ssize_t cor_len _GL_UNUSED)
 {
-	g_return_if_fail (dict);
-	g_return_if_fail (mis);
-	g_return_if_fail (cor);
-
-	if (mis_len < 0)
-		mis_len = strlen (mis);
-
-	if (cor_len < 0)
-		cor_len = strlen (cor);
-
-	g_return_if_fail (mis_len);
-	g_return_if_fail (cor_len);
-
-	g_return_if_fail (g_utf8_validate(mis, mis_len, NULL));
-	g_return_if_fail (g_utf8_validate(cor, cor_len, NULL));
-
-	EnchantSession * session = ((EnchantDictPrivateData*)dict->enchant_private_data)->session;
-	enchant_session_clear_error (session);
-
-	/* if it's not implemented, it's not worth emulating */
-	if (dict->store_replacement)
-		(*dict->store_replacement) (dict, mis, mis_len, cor, cor_len);
 }
 
 void
@@ -1071,7 +1012,7 @@ enchant_broker_request_pwl_dict (EnchantBroker * broker, const char *const pwl)
 }
 
 static EnchantDict *
-_enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
+_enchant_broker_request_dict (EnchantBroker * broker, const char *const tag, const char *const pwl)
 {
 	EnchantDict *dict = (EnchantDict*)g_hash_table_lookup (broker->dict_map, (gpointer) tag);
 	if (dict) {
@@ -1088,7 +1029,7 @@ _enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
 
 			if (dict)
 				{
-					EnchantSession *session = enchant_session_new (provider, tag);
+					EnchantSession *session = enchant_session_new (provider, tag, pwl);
 					EnchantDictPrivateData *enchant_dict_private_data = g_new0 (EnchantDictPrivateData, 1);
 					enchant_dict_private_data->reference_count = 1;
 					enchant_dict_private_data->session = session;
@@ -1103,7 +1044,7 @@ _enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
 }
 
 EnchantDict *
-enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
+enchant_broker_request_dict_with_pwl (EnchantBroker * broker, const char *const tag, const char *const pwl)
 {
 	EnchantDict *dict = NULL;
 
@@ -1115,10 +1056,10 @@ enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
 	char * normalized_tag = enchant_normalize_dictionary_tag (tag);
 	if(!enchant_is_valid_dictionary_tag(normalized_tag))
 		enchant_broker_set_error (broker, "invalid tag character found");
-	else if ((dict = _enchant_broker_request_dict (broker, normalized_tag)) == NULL)
+	else if ((dict = _enchant_broker_request_dict (broker, normalized_tag, pwl)) == NULL)
 		{
 			char * iso_639_only_tag = enchant_iso_639_from_tag (normalized_tag);
-			dict = _enchant_broker_request_dict (broker, iso_639_only_tag);
+			dict = _enchant_broker_request_dict (broker, iso_639_only_tag, pwl);
 			free (iso_639_only_tag);
 		}
 	free (normalized_tag);
@@ -1126,6 +1067,12 @@ enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
 	return dict;
 }
 
+EnchantDict *
+enchant_broker_request_dict (EnchantBroker * broker, const char *const tag)
+{
+	return enchant_broker_request_dict_with_pwl (broker, tag, NULL);
+}
+
 void
 enchant_broker_describe (EnchantBroker * broker, EnchantBrokerDescribeFn fn, void * user_data)
 {
@@ -1416,7 +1363,7 @@ enchant_broker_get_error (EnchantBroker * broker)
 	return broker->error;
 }
 
-char *
+_GL_ATTRIBUTE_MALLOC char *
 enchant_get_user_language(void)
 {
 #if defined(G_OS_WIN32)
