@@ -1,5 +1,6 @@
 /* enchant
  * Copyright (C) 2003 Dom Lachowicz
+ * Copyright (C) 2016-2023 Reuben Thomas
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -16,8 +17,8 @@
  * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  * Boston, MA 02110-1301, USA.
  *
- * In addition, as a special exception, Dom Lachowicz
- * gives permission to link the code of this program with
+ * In addition, as a special exception, the copyright holders
+ * give permission to link the code of this program with
  * non-LGPL Spelling Provider libraries (eg: a MSFT Office
  * spell checker backend) and distribute linked combinations including
  * the two.  You must obey the GNU Lesser General Public License in all
@@ -45,9 +46,6 @@ EnchantPWL* enchant_pwl_init_with_file(const char * file);
 void enchant_pwl_add(EnchantPWL * me, const char *const word, ssize_t len);
 void enchant_pwl_remove(EnchantPWL * me, const char *const word, ssize_t len);
 int enchant_pwl_check(EnchantPWL * me, const char *const word, ssize_t len);
-/*gives the best set of suggestions from pwl that are at least as good as the given suggs*/
-char** enchant_pwl_suggest(EnchantPWL *me, const char *const word,
-			   ssize_t len, char ** suggs, size_t* out_n_suggs);
 void enchant_pwl_free(EnchantPWL* me);
 
 #ifdef __cplusplus
